unit ExifTool;

interface
uses Classes, StdCtrls;

type
  ET_OptionsRec=record     //don't define '-a' (because of Filelist custom columns)
    ETLangDef: string;     //don't define '-e' (because of extracting previews)
    ETBackupMode: string;
    ETFileDate: string;
    ETSeparator: string;
    ETMinorError: string;
    ETGpsFormat: string;
    ETShowNumber: string;
    //ETuseLang: boolean;
  end;

const
  CRLF=#13#10;
var
  ETout, ETerr: TStringList;
  ETOpenReady: boolean=false;
  ETCounterLabel: TLabel=nil;
  ETCounter: smallint=0;
  ET_Options: ET_OptionsRec=
   (ETLangDef: '';
    ETBackupMode:'-overwrite_original'+CRLF; //or ='' or='overwrite_original_in_place'
    ETFileDate:'';                           //or '-P'+CRLF (preserve FileModify date)
    ETSeparator:'-sep'+CRLF+'*'+CRLF;        //or '' (for keywords etc.)
    ETMinorError:'';                         //or '-m'+CRLF
    ETGPSformat:'-c'+CRLF+'%d%.4f'+CRLF;    //or '-c'+CRLF+'%.6f'+CRLF (for decimal)
    ETShowNumber:'');                         //or '-n'+CRLF
//    ETuseLang: false);

function ET_StayOpen(WorkDir:string): boolean;
function ET_OpenExec(ETcmd:string; FNames:AnsiString=''; TagsFromFile:boolean=false): boolean;
procedure ET_OpenExit;
function ExecET(ETcmd,Fnames,WorkDir:AnsiString; UseUtf8:boolean=true): boolean;
function ExecCMD(xCmd,WorkDir:AnsiString): boolean;

implementation
uses Main, Windows, Forms, SysUtils;

const
  szPipeBuffer=65535;

var
  ETstream: TMemoryStream;
  ETprocessInfo: TProcessInformation;
  PipeBuffer: array[0..szPipeBuffer] of Byte;
  PipeInRead, PipeInWrite: THandle;
  PipeOutRead, PipeOutWrite: THandle;
  PipeErrRead, PipeErrWrite: THandle;
  ETrunning: boolean=false;
  ETShowCounter: boolean=false;
//============================== ET_Open mode ==================================
function ET_StayOpen(WorkDir:string): boolean;
var
  SecurityAttr: TSecurityAttributes;
  StartupInfo: TStartupInfo;
  PWorkDir: PChar;
  ETcmd: WideString;
begin
  ETcmd:='exiftool -stay_open True -@ -';
  if ETrunning then ET_OpenExit; //-changing WorkDir OnTheFly requires Exit first
  FillChar(ETprocessInfo, SizeOf(TProcessInformation), #0);
  FillChar(SecurityAttr, SizeOf(TSecurityAttributes), #0);
  SecurityAttr.nLength:=SizeOf(SecurityAttr);
  SecurityAttr.bInheritHandle:=true;
  SecurityAttr.lpSecurityDescriptor:=nil;
  CreatePipe(PipeInRead, PipeInWrite, @SecurityAttr, 0);
  CreatePipe(PipeOutRead, PipeOutWrite, @SecurityAttr, 65535);
  CreatePipe(PipeErrRead, PipeErrWrite, @SecurityAttr, 32767);
  FillChar(StartupInfo, SizeOf(TStartupInfo), #0);
  StartupInfo.cb:=SizeOf(StartupInfo);
  with StartupInfo do begin
    hStdInput:=PipeInRead;
    hStdOutput:=PipeOutWrite;
    hStdError:=PipeErrWrite;
    wShowWindow:=SW_HIDE;
    dwFlags:=STARTF_USESHOWWINDOW or STARTF_USESTDHANDLES;
  end;
  if WorkDir='' then PWorkDir:=nil else PWorkDir:=PChar(WorkDir);
  if CreateProcess(nil, PChar(ETcmd), nil, nil, true,
  CREATE_DEFAULT_ERROR_MODE or CREATE_NEW_CONSOLE or NORMAL_PRIORITY_CLASS,
  nil, PWorkDir, StartupInfo, ETprocessInfo) then begin
    ETrunning:=true;
    CloseHandle(PipeInRead);
    CloseHandle(PipeOutWrite);
    CloseHandle(PipeErrWrite);
  end else begin
    ETrunning:=false;
    CloseHandle(PipeInRead); CloseHandle(PipeInWrite);
    CloseHandle(PipeOutRead); CloseHandle(PipeOutWrite);
    CloseHandle(PipeErrRead); CloseHandle(PipeErrWrite);
  end;
  ETOpenReady:=ETrunning;
  result:=ETrunning;
end;

function ET_OpenExec(ETcmd:string; Fnames:AnsiString=''; TagsFromFile:boolean=false): boolean;
// -Fnames must be of AnsiString type here, otherwise some foreign
// characters in filenames might no be recognized by ExifTool!
var
  ETprm: AnsiString;
  FinalCmd: AnsiString;
  BytesCount: Dword;
  i: word;
  BuffContent: ^word;
  BuffAddress: ^dword;
  EndReady: boolean;
begin
  result:=false;
  ETout.Clear; ETerr.Clear;
  if ETrunning and ETOpenReady and (Length(ETcmd)>1) then begin
    ETOpenReady:=false;

    if pos('||', ETcmd)>0 then
      ETcmd:=StringReplace(ETcmd, '||', CRLF, [rfReplaceAll]);
    ETcmd:=ETcmd+CRLF;

    ETprm:='-v0'+CRLF;  //-for file counter!
    with ET_Options do begin
      if ETLangDef<>'' then ETprm:=ETprm+'-lang'+CRLF+ETLangDef;
      ETprm:=ETprm+ETBackupMode;
      ETprm:=ETprm+ETSeparator;
      ETprm:=ETprm+ETMinorError+ETFileDate;
      ETprm:=ETprm+ETGPSformat+ETShowNumber;
      //+further options...
    end;

    //Only Utf8encode input parameters -but not if
    // TagsFromFile is used, because after TagsFromFile,
    // there can be a filename parameter:
    if TagsFromFile then FinalCmd:=ETprm+ETcmd
    else FinalCmd:=Utf8Encode(ETprm+ETcmd);

    //Add Ansi filenames:
    if FNames<>'' then FinalCmd:=FinalCmd+FNames+CRLF;
    FinalCmd:=FinalCmd+'-execute9'+CRLF;

    ETShowCounter:=(ETCounterLabel<>nil) and (ETCounter>1);
    if ETShowCounter then begin
      ETCounterLabel.Visible:=true;
      FMain.ShellList.AutomaticRefresh:=false;
    end;

    WriteFile(PipeInWrite, FinalCmd[1], Length(FinalCmd), BytesCount, nil);

    Screen.Cursor:=-11; //crHourglass
    FlushFileBuffers(PipeInWrite);
    //========= Read StdOut =======================
    ETstream.Clear;
    BuffAddress:=Addr(BuffContent);
    EndReady:=false;
    repeat
      ReadFile(PipeOutRead, PipeBuffer, szPipeBuffer, BytesCount, nil);
      BuffContent:=@PipeBuffer[0];
      if BytesCount>1 then begin
        if BuffContent^=$3D3D then begin
          Dec(ETcounter);
          if ETShowCounter then ETCounterLabel.Caption:=IntToStr(ETCounter);
        end else ETstream.Write(PipeBuffer, BytesCount);
        for i:=0 to BytesCount-1 do begin
          if BuffContent^=$7D39 then EndReady:=true; // line contains '9}'
          inc(BuffAddress^,1);
        end;
      end;
      Application.ProcessMessages;
    until EndReady;
    ETstream.Position:=0; ETout.LoadFromStream(ETstream);
    //========= Read StdErr =======================
    ETstream.Clear;  i:=0;
    repeat
      PeekNamedPipe(PipeErrRead, nil, 0, nil, @BytesCount, nil);
      if BytesCount>0 then begin
        ReadFile(PipeErrRead, PipeBuffer, szPipeBuffer, BytesCount, nil);
        ETstream.Write(PipeBuffer, BytesCount);
      end;
      inc(i);
    until (BytesCount=0) or (i>5); // max 2 attempts observed
    ETstream.Position:=0; ETerr.LoadFromStream(ETstream);
    //========= Cleanup ETout ======================
    i:=0;
    while i<ETout.Count do begin
      //no difference observed between UTF8ToString() and UTF8ToAnsi()
      ETout[i]:=UTF8ToAnsi(ETout[i]);
      if pos('{r',ETout[i])>0 then ETout.Delete(i) //delete '{ready..}' lines
                              else inc(i);
    end;
    //----------------------------------------------
    if ETShowCounter then ETCounterLabel.Visible:=false;
    ETCounter:=0;
    FMain.ShellList.AutomaticRefresh:=true;
    result:=true; Screen.Cursor:=0; //crDefault
    ETOpenReady:=true;
  end;
end;

procedure ET_OpenExit;
const
  ExitCmd:Utf8string='-stay_open'+CRLF+'False'+CRLF;
var
  BytesCount: dWord;
begin
  if ETrunning then begin
    WriteFile(PipeInWrite, ExitCmd[1], Length(ExitCmd), BytesCount, nil);
    FlushFileBuffers(PipeInWrite);
    CloseHandle(ETprocessInfo.hThread);
    CloseHandle(ETprocessInfo.hProcess);
    CloseHandle(PipeInWrite);
    CloseHandle(PipeOutRead);
    CloseHandle(PipeErrRead);
    ETrunning:=false;
  end;
end;
//^^^^^^^^^^^^^^^^^^^^^^^^^^^ End of ET_Open mode ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
//
//=========================== ET classic mode ==================================
function ExecET(ETcmd,Fnames,WorkDir:AnsiString; UseUtf8:boolean=true): boolean;
var
  ProcessInfo: TProcessInformation;
  SecurityAttr: TSecurityAttributes;
  StartupInfo: TStartupInfoA;
  PipeOutRead, PipeOutWrite: THandle;
  PipeErrRead, PipeErrWrite: THandle;
  ETprm: AnsiString;
  PWorkDir: PAnsiChar;
  BytesCount: DWord;
  i: word;
  BuffContent: ^word;
begin
  FillChar(ProcessInfo, SizeOf(TProcessInformation), #0);
  FillChar(SecurityAttr, SizeOf(TSecurityAttributes), #0);
  SecurityAttr.nLength:=SizeOf(SecurityAttr);
  SecurityAttr.bInheritHandle:=true;
  SecurityAttr.lpSecurityDescriptor:=nil;
  CreatePipe(PipeOutRead, PipeOutWrite, @SecurityAttr, 65535);
  CreatePipe(PipeErrRead, PipeErrWrite, @SecurityAttr, 32767);
  FillChar(StartupInfo, SizeOf(TStartupInfo), #0);
  StartupInfo.cb:=SizeOf(StartupInfo);
  with StartupInfo do begin
    hStdInput:=0;
    hStdOutput:=PipeOutWrite;
    hStdError:=PipeErrWrite;
    wShowWindow:=SW_HIDE;
    dwFlags:=STARTF_USESHOWWINDOW or STARTF_USESTDHANDLES;
  end;
  if WorkDir='' then PWorkDir:=nil else PWorkDir:=PAnsiChar(WorkDir);

  ETprm:='exiftool -v0 ';
  with ET_Options do begin
    if ETLangDef<>'' then ETprm:=ETprm+'-lang'+CRLF+ETLangDef;
    ETprm:=ETprm+ETBackupMode;
    ETprm:=ETprm+ETSeparator+ETMinorError;
    ETprm:=ETprm+ETFileDate+ETGpsFormat;
  end;
  ETprm:=StringReplace(ETprm, CRLF, ' ', [rfReplaceAll]);

  if UseUTF8 then ETcmd:=ETprm+Utf8encode(ETcmd)+Fnames
             else ETcmd:=ETprm+ETcmd+Fnames;
  ETShowCounter:=(ETCounterLabel<>nil) and (ETCounter>1);
  if ETShowCounter then begin
    ETCounterLabel.Visible:=true;
    FMain.ShellList.AutomaticRefresh:=false;
  end;

  ETout.Clear; ETerr.Clear;
  Screen.Cursor:=-11; //=crHourGlass
  if CreateProcessA(nil, PAnsiChar(ETcmd), nil, nil, true,
    CREATE_DEFAULT_ERROR_MODE or CREATE_NEW_CONSOLE or NORMAL_PRIORITY_CLASS,
    nil, PWorkDir, StartupInfo, ProcessInfo)
  then begin
    result:=true;
    CloseHandle(PipeOutWrite);
    CloseHandle(PipeErrWrite);
  end else begin
    result:=false;
    CloseHandle(PipeOutRead); CloseHandle(PipeOutWrite);
    CloseHandle(PipeErrRead); CloseHandle(PipeErrWrite);
  end;
  if result then begin
    //========= Read StdOut =======================
    ETstream.Clear;  //BuffAddress:=Addr(BuffContent);
    repeat
      ReadFile(PipeOutRead, PipeBuffer, szPipeBuffer, BytesCount, nil);
      BuffContent:=@PipeBuffer[0];
      if BytesCount>1 then begin
        if BuffContent^=$3D3D then begin  // line starts with '=='
          Dec(ETCounter);
          if ETShowCounter then ETCounterLabel.Caption:=IntToStr(ETCounter);
        end; // else ETstream.Write(PipeBuffer, BytesCount);
        ETstream.Write(PipeBuffer, BytesCount);
      end;
      Application.ProcessMessages;
    until BytesCount=0;
    ETstream.Position:=0; ETout.LoadFromStream(ETstream);
    CloseHandle(PipeOutRead);
    //========= Read StdErr =======================
    ETstream.Clear;
    repeat
      ReadFile(PipeErrRead, PipeBuffer, szPipeBuffer, BytesCount, nil);
      ETstream.Write(PipeBuffer, BytesCount);
    until (BytesCount=0);
    ETstream.Position:=0; ETerr.LoadFromStream(ETstream);
    CloseHandle(PipeErrRead);
    //========= Cleanup ETout ======================
    if UseUtf8 then begin
      i:=0;
      while i<ETout.Count do begin
        ETout[i]:=Utf8ToAnsi(ETout[i]);
        inc(i);
      end;
    end;
    //----------------------------------------------
    WaitForSingleObject(ProcessInfo.hProcess,5000); //msec=5sec /or INFINITE
    CloseHandle(ProcessInfo.hThread); CloseHandle(ProcessInfo.hProcess);
    result:=true;
  end;
  if ETShowCounter then ETCounterLabel.Visible:=false;
  ETCounter:=0;
  FMain.ShellList.AutomaticRefresh:=true;
  Screen.Cursor:=0; //crDefault
end;
//^^^^^^^^^^^^^^^^^^^^^^^^^^ End of ET Classic mode ^^^^^^^^^^^^^^^^^^^^^^^^^^^^
//
//==============================================================================
function ExecCMD(xCmd,WorkDir:AnsiString): boolean;
var
  ProcessInfo: TProcessInformation;
  SecurityAttr: TSecurityAttributes;
  StartupInfo: TStartupInfoA;
  PipeOutRead, PipeOutWrite: THandle;
  PipeErrRead, PipeErrWrite: THandle;
  PWorkDir: PAnsiChar;
  BytesCount: DWord;
begin
  FillChar(ProcessInfo, SizeOf(TProcessInformation), #0);
  FillChar(SecurityAttr, SizeOf(TSecurityAttributes), #0);
  SecurityAttr.nLength:=SizeOf(SecurityAttr);
  SecurityAttr.bInheritHandle:=true;
  SecurityAttr.lpSecurityDescriptor:=nil;
  CreatePipe(PipeOutRead, PipeOutWrite, @SecurityAttr, 8192);
  CreatePipe(PipeErrRead, PipeErrWrite, @SecurityAttr, 8192);
  FillChar(StartupInfo, SizeOf(TStartupInfo), #0);
  StartupInfo.cb:=SizeOf(StartupInfo);
  with StartupInfo do begin
    hStdInput:=0;
    hStdOutput:=PipeOutWrite;
    hStdError:=PipeErrWrite;
    wShowWindow:=SW_HIDE;
    dwFlags:=STARTF_USESHOWWINDOW or STARTF_USESTDHANDLES;
  end;
  if WorkDir='' then PWorkDir:=nil else PWorkDir:=PAnsiChar(WorkDir);

  ETout.Clear; ETerr.Clear;
  Screen.Cursor:=-11; //=crHourGlass
  if CreateProcessA(nil, PAnsiChar(xCmd), nil, nil, true,
    CREATE_DEFAULT_ERROR_MODE or CREATE_NEW_CONSOLE or NORMAL_PRIORITY_CLASS,
    nil, PWorkDir, StartupInfo, ProcessInfo)
  then begin
    result:=true;
    CloseHandle(PipeOutWrite);
    CloseHandle(PipeErrWrite);
  end else begin
    result:=false;
    CloseHandle(PipeOutRead); CloseHandle(PipeOutWrite);
    CloseHandle(PipeErrRead); CloseHandle(PipeErrWrite);
  end;
  if result then begin
    //========= Read StdOut =======================
    ETstream.Clear;  //BuffAddress:=Addr(BuffContent);
    repeat
      ReadFile(PipeOutRead, PipeBuffer, szPipeBuffer, BytesCount, nil);
      if BytesCount>1 then ETstream.Write(PipeBuffer, BytesCount);
      Application.ProcessMessages;
    until BytesCount=0;
    ETstream.Position:=0; ETout.LoadFromStream(ETstream);
    CloseHandle(PipeOutRead);
    //========= Read StdErr =======================
    ETstream.Clear;
    repeat
      ReadFile(PipeErrRead, PipeBuffer, szPipeBuffer, BytesCount, nil);
      ETstream.Write(PipeBuffer, BytesCount);
    until (BytesCount=0);
    ETstream.Position:=0; ETerr.LoadFromStream(ETstream);
    CloseHandle(PipeErrRead);
    //----------------------------------------------
    WaitForSingleObject(ProcessInfo.hProcess,5000); //msec=5sec /or INFINITE
    CloseHandle(ProcessInfo.hThread); CloseHandle(ProcessInfo.hProcess);
    result:=true;
  end;
  Screen.Cursor:=0; //crDefault
end;
//==============================================================================

initialization
begin
  ETout:=TStringList.Create; ETerr:=TStringList.Create;
  ETstream:=TMemoryStream.Create;
  ETCounterLabel:=nil;
end;

finalization
begin
  ET_OpenExit;
  ETstream.Free;
  ETout.Free; ETerr.Free;
end;


end.
